from collections import defaultdict
import json
import pathlib
from typing import Callable, DefaultDict, Dict, Iterable, List, Literal, Optional, Union

from networkx import descendants
from crtl.components.ann_vector_kb import MentionCandidate
from crtl.components.crtl_el import CRTLTextLinker, get_setters_and_getters
from crtl.lkb.lkb import Concept, LexicalKnowledgeBase
from spacy.tokens import Doc, Span, SpanGroup
from spacy.language import Language
from spacy.util import registry
from spacy.pipeline import Pipe
from spacy.training import Example
from functools import reduce


class DiscrepancyRectifier:
    def __init__(
        self,
        cui_to_cer_label_mapping: dict[str, list[str]],
        low_to_high_map: dict[Union[str, int], List[Union[str, int]]],
    ):
        self._cui_to_cer_label_mapping = cui_to_cer_label_mapping
        self.low_to_high_map = low_to_high_map

    @classmethod
    def initialize(
        cls,
        lkb: LexicalKnowledgeBase,
        cer_label_to_cuis: Dict[str, list[str]],  # cui-to-cer_label mapping)
        hierarchical_relations: List[str],
    ):

        cui_to_cer_label_mapping = defaultdict(lambda: [])
        for cer_label, cui_list in cer_label_to_cuis.items():
            for cui in cui_list:
                cui_to_cer_label_mapping[cui].append(cer_label)
        low_to_high_map = cls._generate_low_to_high_mapping(
            lkb, cui_to_cer_label_mapping, hierarchical_relations
        )
        return cls(dict(cui_to_cer_label_mapping), dict(low_to_high_map))

    def get_filtered_mention_candidates(
        self, label, mention_candidates
    ) -> List[MentionCandidate]:
        new_mention_candidates = []
        for mention_candidate in mention_candidates:
            mention_candidate: MentionCandidate
            if label in self._get_high_level_cer_labels(mention_candidate.concept_id):
                new_mention_candidates.append(mention_candidate)

        return new_mention_candidates

    def _get_high_level_cer_labels(self, cui) -> List[str]:
        cer_labels = []
        if cui in self.low_to_high_map:
            for high_level_cui in self.low_to_high_map[cui]:
                cer_labels.extend(self._cui_to_cer_label_mapping[high_level_cui])
        return cer_labels

    @classmethod
    def _generate_low_to_high_mapping(
        cls,
        lkb: LexicalKnowledgeBase,
        cui_to_cer_label_mapping: Dict[str, list[str]],
        hierarchical_relations: List[str],
    ) -> DefaultDict[Union[str, int], List[Union[str, int]]]:
        # Function to collect the nodes in a breadth-first traversal
        dag = lkb.create_dag(hierarchical_relations)
        low_to_high_map = defaultdict(lambda: [])
        for high_level_concept_id in cui_to_cer_label_mapping.keys():
            low_to_high_map[high_level_concept_id].append(high_level_concept_id)
            for descendant in descendants(dag, high_level_concept_id):
                low_to_high_map[descendant].append(high_level_concept_id)
        return low_to_high_map

    def to_disk(self, path):
        if not path.exists():
            path.mkdir()
        with open(path / "high_level_kb_id_to_name.json", "w+") as f:
            json.dump(self._cui_to_cer_label_mapping, f)
        with open(path / "low_to_high_map.json", "w+") as f:
            json.dump(self.low_to_high_map, f)

    @classmethod
    def from_disk(cls, path):
        with open(path / "high_level_kb_id_to_name.json") as f:
            cui_to_cer_label_mapping = json.load(f)
        with open(path / "low_to_high_map.json") as f:
            low_to_high_map = json.load(f)
        return cls(cui_to_cer_label_mapping, low_to_high_map)


class CandidateSelector(Pipe):
    def __init__(
        self,
        threshold: float,
        ent_or_span: Literal["ent", "span"],
        span_key: Optional[str] = None,
        scorer: Optional[Callable] = None,
        cer_label_subset: Optional[List[str]] = None,
        hierarcical_relations: Optional[List[str]] = None,
    ):
        self.scorer = scorer
        self._threshold = threshold
        self._hierarchical_relations = hierarcical_relations
        self.cer_label_subset = cer_label_subset
        self.span_getter, self.span_setter = get_setters_and_getters(
            ent_or_span, span_key
        )

    def _select_candidate(
        self, span, mention_candidates: Iterable[MentionCandidate]
    ) -> str:
        if not self.cer_label_subset or span.label_ in self.cer_label_subset:
            for candidate in mention_candidates:

                if candidate.similarity > self._threshold:
                    return str(candidate.concept_id)
            return CRTLTextLinker.NIL
        else:
            return CRTLTextLinker.NA

    def __call__(self, doc: Doc) -> Doc:
        spans = self.span_getter(doc)
        new_spans = []
        for span in spans:
            if self.cer_label_subset:
                mention_candidates = (
                    self._discrepancy_rectifier.get_filtered_mention_candidates(
                        span.label_, span._.candidate_mentions
                    )
                )
            else:
                mention_candidates = span._.candidate_mentions
            kb_id = self._select_candidate(span, mention_candidates)
            new_spans.append(
                Span(doc, span.start, span.end, label=span.label_, kb_id=kb_id)
            )
        self.span_setter(doc, new_spans)
        return doc

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        lkb_getter: Callable[[], LexicalKnowledgeBase],
        cer_label_to_cuis: Optional[
            Dict[str, list[str]]
        ] = None,  # cui-to-cer_label mapping
        *,
        nlp=None,
    ):
        lkb = lkb_getter()
        if cer_label_to_cuis:
            assert self._hierarchical_relations
            self._discrepancy_rectifier = DiscrepancyRectifier.initialize(
                lkb, cer_label_to_cuis, self._hierarchical_relations
            )
        else:
            self._discrepancy_rectifier = None

    def to_disk(self, path, exclude=tuple()):
        if not path.exists():
            path.mkdir()
        if self.cer_label_subset:
            self._discrepancy_rectifier.to_disk(path / "discrepancy_rectifier")

    def from_disk(self, path, exclude=tuple()):
        if self.cer_label_subset:
            self._discrepancy_rectifier = DiscrepancyRectifier.from_disk(
                path / "discrepancy_rectifier"
            )


@Language.factory(
    name="basic_candidate_selector",
    default_config={
        "threshold": 0.7,
        "ent_or_span": "ent",
        "span_key": None,
        "scorer": {
            "@scorers": "crtl.score_links",
            "ent_or_span": "ent",
        },
        "cer_label_subset": None,
        "hierarchical_relations": None,
    },
    default_score_weights={
        "strong_nel_ete_f": 1.0,
        "strong_nel_ete_r": None,
        "strong_nel_ete_p": None,
    },
)
def make_basic_candidate_selector(
    nlp: Language,
    name: str,
    threshold: float,
    ent_or_span: Literal["ent", "span"],
    span_key: Optional[str],
    scorer: Optional[Callable] = None,
    cer_label_subset: Optional[List[str]] = None,
    hierarchical_relations: Optional[List[str]] = None,
):
    return CandidateSelector(
        threshold,
        ent_or_span,
        span_key,
        scorer,
        cer_label_subset,
        hierarchical_relations,
    )


@registry.readers("crtl.read_concept_ids_from_json")
def make_high_level_concept_id_reader(
    path: Union[str, pathlib.Path]
) -> Dict[str, list[str]]:
    with open(path) as f:
        return json.load(f)
