#!usr/bin/python3 python3


import sys

sys.path.append("../../../")

import spacy
from spacy.tokens import Doc, Span
from spacy.training import Example
from spacy.util import minibatch, compounding

from sentence_transformers.readers import InputExample

import torch
from torch.utils.data import DataLoader

import random
from pathlib import Path

from crtl.lkb.lkb import LexicalKnowledgeBase

from entity_linker import EntityLinker, EncoderPredictions, MarginMSELoss
from document import Document
from embedding_utils import EmbeddingUtils
from distancer import Distancer
from dual_encoder_pipe import DualEncoderPipe
from encoder import Encoder


from typing import Iterable


def train(train_docs: Iterable[Example], num_iterations: int):
    Doc.set_extension("embedding", default=None)
    Span.set_extension("linked_entity", default=None)
    Span.set_extension("entity_score", default=None)

    nlp = spacy.blank("en")

    nlp.add_pipe("parser")
    nlp.add_pipe("sentencizer")
    nlp.add_pipe("dual_encoder", last=True)

    other_pipes = [pipe for pipe in nlp.pipe_names if pipe != "dual_encoder"]

    with nlp.disable_pipes(*other_pipes):
        optimizer = nlp.resume_training()
        for i in range(num_iterations):
            random.shuffle(train_docs)
            losses = {}
            for batch in minibatch(train_docs, size=compounding(4.0, 32.0, 1.001)):
                texts, annotations = zip(*batch)
                nlp.update(texts, annotations, drop=0.15, sgd=optimizer, losses=losses)
            if i % 100 == 0:
                print("iteration: {} losses: {}".format(i, losses))
        nlp.to_disk("/de-cel")


def test_loss():
    Doc.set_extension("embedding", default=None)
    Span.set_extension("linked_entity", default=None)
    Span.set_extension("entity_score", default=None)

    loss = MarginMSELoss(pool_negatives=True)

    pred = torch.Tensor([1.0, 0.0, 0.0])
    pred.requires_grad = True
    positive = torch.Tensor([0.0, 0.7, 0.3])
    negatives = [torch.Tensor([1.0, 0.0, 0.0]), torch.Tensor([0.7, 0.0, 0.3])]

    output = loss(pred, positive, negatives)
    output.backward()

    print("pred: ", pred)
    print("positive: ", positive)
    print("negative: ", negatives)
    print("output: ", output)


def test():
    Doc.set_extension("embedding", default=None)
    Span.set_extension("linked_entity", default=None)
    Span.set_extension("entity_score", default=None)

    el = EntityLinker(
        LexicalKnowledgeBase.from_disk(
            "/Users/joshuatint/Developer/CRTL/src/crtl/components/dual_encoder/test_data/LKBTest.jsonl"
        ),
        "entities",
        0.5,
    )

    training_loader = DataLoader(
        [
            ("schnozz", "nose", ["ear", "mouth", "throat"]),
            (
                "I was coughing this morning",
                "cough",
                ["sneeze", "cough drops", "diabetes", "cancer"],
            ),
            (
                "I have about 3 palpable lymph nodes",
                "lymph node",
                ["neck", "cancer", "Hodgkin's lymphoma"],
            ),
            ("One is on my neck", "neck", ["scar tissue", "dietician", "skin"]),
            (
                "The ENT brushed it off",
                "ear nose and throat specialist",
                ["clinician", "nurse", "oncologist", "endometriosis"],
            ),
            (
                "X-ray came out normal",
                "X-ray scan",
                ["MRI Scan", "pediatrics", "bone density"],
            ),
        ]
        * 100
    )

    validation_loader = DataLoader(
        [
            (
                "common cold",
                "rhinovirus",
                ["cold temperature", "chills", "osteoporosis"],
            ),
            ("CT scan", "CAT scan", ["MRI scan", "common transmission CT", "nose"]),
        ]
    )

    el.train(training_loader, validation_loader)


def end_to_end_test():
    Doc.set_extension("embedding", default=None)
    Span.set_extension("linked_entity", default=None)
    Span.set_extension("entity_score", default=None)

    nlp = spacy.blank("en")

    nlp.add_pipe("parser")
    nlp.add_pipe("sentencizer")
    nlp.add_pipe(
        "dual_encoder",
        config={
            "lkb_path_str": "/Users/joshuatint/Developer/CRTL/src/crtl/components/dual_encoder/test_data/LKBTest.jsonl"
        },
        last=True,
    )

    text = "I have Renal Disease. Dang it! I also think I have Prader-Willi :("

    nlp.initialize()

    doc = nlp(text)

    print(doc.to_json())
    print(doc)
    for span in doc.spans["entities"]:
        print("{}: {}".format(span, span._.linked_entity.canonical_name))


def encoder_test():
    Doc.set_extension("embedding", default=None)
    nlp = spacy.blank("en")
    # encoder_pipe = EncoderPipe(nlp, "encoder")

    encoder = Encoder()

    # print("Text Encoding:", doc._.embedding)

    ice_cream = nlp("I like ice cream")
    cake = nlp("I like cake")
    woodshop = nlp("I took woodshop in high school")
    engineering = nlp("I liked my high school engineering class")

    print(
        "expected: low, got: {}".format(
            Distancer.distance_between(encoder(ice_cream.text), encoder(cake.text))
        )
    )
    print(
        "expected: low, got: {}".format(
            Distancer.distance_between(
                encoder(engineering.text), encoder(woodshop.text)
            )
        )
    )
    print(
        "expected: high, got: {}".format(
            Distancer.distance_between(encoder(engineering.text), encoder(cake.text))
        )
    )
    print(
        "expected: high, got: {}".format(
            Distancer.distance_between(encoder(ice_cream.text), encoder(woodshop.text))
        )
    )

    """
    config_encoder = BertConfig()

    config_decoder = BertConfig()

    config = EncoderDecoderConfig.from_encoder_decoder_configs(config_encoder, config_decoder)   
    model = EncoderDecoderModel(config=config)
    tokenizer = AutoTokenizer.from_pretrained("bert-base-cased")
    transformer_pipeline = pipeline('transformer', model=model, tokenizer=tokenizer)
    text = "my text for named entity recognition here."
    print(transformer_pipeline(text))
    """


def main():
    # test()
    # encoder_test()
    # train_encoder()
    end_to_end_test()
    """
    train(train_docs, 500)
    """


"""def pretrained_test():
    nlp = spacy.load("en_core_web_sm")

    (text, labels) = politics_test

    el = EntityLinker(
        #DualEncoder(),
        labels,
    )

    doc = Document(nlp(text))

    el.link_entities_in_document(doc, show_progress_bar=True)

    print(doc.get_annotated_string(True))
"""


"""
    avged_embedding = EmbeddingUtils.average_embeddings(
        [
            el.dual_encoder.encode("it"),
            el.dual_encoder.encode("of"),
            el.dual_encoder.encode("the"),
            el.dual_encoder.encode("not"),
            el.dual_encoder.encode("so"),
            el.dual_encoder.encode("could"),
        ]
    )

    print(avged_embedding)
"""


if __name__ == "__main__":
    main()
