import spacy
from spacy import Language
from spacy.training import Example
from spacy.pipeline import TrainablePipe
from spacy.tokens import Span, Doc
from spacy.vocab import Vocab

from pathlib import Path

from crtl.lkb.lkb import LexicalKnowledgeBase

from embedding_utils import EmbeddingUtils
from entity_linker import EntityLinker
from recognizer import EntityRecognizer

from typing import Callable, Optional, Iterable, List


class DualEncoderPipe(TrainablePipe):
    """
    This class is the trainable pipe that represents the end-to-end recognition and linking process for a document

    vocab: Vocab
        The vocab of the lanugage in use
    name: str
        The name of the pipe component
    lkb: LexicalKnowledgeBase
        The knowledge base of concepts to link
    entity_recognizer: EntityRecognizer
        The entity recognizer component
    entity_linker: EntityLinker
        The linker component
    """

    def __init__(self, vocab: Vocab, name: str, **cfg):
        """
        The config should contain the following:

        lkb: LexicalKnowledgeBase
            The lexical knowledge base whose concepts are linked in the document.
        """
        # TrainablePipe.__init__(self, vocab, None, name, cfg)
        self.vocab = vocab
        self.name = name
        self.lkb: LexicalKnowledgeBase = cfg["lkb"]
        self.entity_recognizer = EntityRecognizer(key="entities")
        self.entity_linker = EntityLinker(self.lkb, span_key="entities", min_score=0.5)

    def __call__(self, doc: Doc, verbose: bool = False):
        """
        __call__

        Annotate the document with a list of entities that are linked to concepts in the pipe's lkb

        -------
        PARAMS:

        doc: Doc
            The SpaCy doc to annotate
        verbose: bool
            If true, prints out report messages when spans are dropped
        """

        doc = self.entity_recognizer(doc)
        linked_doc = self.entity_linker(doc)
        filtered = linked_doc.spans[self.entity_linker.span_key]

        # Remove spans that couldn't be linked to any concept
        cleaned_spans = [span for span in filtered if span._.linked_entity is not None]

        if verbose:
            for span in filtered:
                if span not in cleaned_spans:
                    print("Dropping span {} due to lack of link".format(span))

        # Replace the generic "relevant" label with the name of the link (Possibly change this to ID)
        for span in cleaned_spans:
            span.label_ = span._.linked_entity.canonical_name
        linked_doc.spans[self.entity_linker.span_key] = cleaned_spans

        return linked_doc

    @property
    def is_trainable():
        return False

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        *,
        nlp: Language,
    ):
        self.entity_linker.initialize_candidate_embeddings()

    def predict(self, docs):
        for doc in docs:
            for span in doc.spans:
                span._.embedding = self(span.text)


@Language.factory(name="dual_encoder")
def make_dual_encoder_pipe(nlp: Language, name: str, lkb_path_str: str):
    """
    make_dual_encoder_pipe

    The Language Factory function that SpaCy uses to create the dual encoder.
    """
    return DualEncoderPipe(
        nlp.vocab,
        name,
        lkb=LexicalKnowledgeBase.from_disk(Path(lkb_path_str)),
    )
