from typing import List, Iterable, Tuple, Callable, Optional, Dict

from spacy import Language
from spacy.tokens import Span, Doc
from spacy.pipeline import TrainablePipe
from spacy.training import Example
from spacy.vocab import Vocab
from thinc.api import Model

import torch
from torch import Tensor
import torch.nn.functional as F
from torch.utils.data import DataLoader

from transformers import AutoTokenizer, AutoModel
from sentence_transformers import losses
from sentence_transformers.util import dot_score


class Encoder:
    """
    This class is responsible for computing and returning the embedding of a piece of text. It can
    be trained seperately as a mention or candidate encoder

    ---------
    PARAMS:

    model: BertModel
        The model used to compute embeddings
    tokenizer: Tokenizer
        The tokenizer for the model
    loss: MultipleNegativesRankingLoss
        The loss function object for model training
    """

    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(
            "sentence-transformers/all-MiniLM-L6-v2"
        )
        self.model = AutoModel.from_pretrained("sentence-transformers/all-MiniLM-L6-v2")
        self.loss = losses.MultipleNegativesRankingLoss(
            model=self.model, scale=1.0, similarity_fct=dot_score
        )

    def __call__(self, text: str) -> torch.Tensor:
        """
        __call__

        Return the model's embedding of a piece of text

        ---------
        PARAMS:

        text: str
            The text to compute the embedding of
        """

        encoded_input = self.tokenizer(text, return_tensors="pt")
        input_ids = encoded_input["input_ids"]

        # Extract and apply attention mask:
        attention_mask = encoded_input["attention_mask"]
        # with torch.no_grad():
        outputs = self.model(input_ids=input_ids, attention_mask=attention_mask)

        # Perform pooling
        sentence_embeddings = self.mean_pooling(outputs, attention_mask)

        # Normalize embeddings
        sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)

        return sentence_embeddings.squeeze()

    def mean_pooling(self, model_output, attention_mask):
        """
        mean_pooling

        Applies a mean pooling to the last hidden layer of the underlying model using an attention mast

        ------
        PARAMS:

        model_output: BaseModelOutputWithPoolingAndCrossAttentions
            The output of the model after being fed input id's

        attention_mask: Tensor
            The attention mask for the token embeddings
        """

        # First element of model_output contains all token embeddings
        token_embeddings = model_output[0]

        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
            input_mask_expanded.sum(1), min=1e-9
        )

    def get_loss(
        self, sentence_features: Iterable[Dict[str, Tensor]], labels: Tensor
    ) -> float:
        """
        get_loss

        Returns the result of the model's loss function at this point during training

        ------
        PARAMS:

        sentence_features: Iterable[Dict[str, Tensor]]
            The list of sentence features for the training result

        labels: Tensor
            The tensor of labels for the model
        """

        return self.loss.forward(sentence_features, labels)
