from typing import List, Callable, Iterable, Dict, Tuple, NamedTuple, Optional


from spacy.tokens import Span, Doc
from spacy.vocab import Vocab
from crtl.lkb.lkb import Concept, LexicalKnowledgeBase

import torch
from torch.utils.tensorboard import SummaryWriter
from torch.utils.data import DataLoader

from datetime import datetime

from distancer import Distancer
from context_window import ContextWindow

from encoder import Encoder


class EncoderPredictions(NamedTuple):
    """
    EncoderPredictions

    Holds a set of embeddings produced by the encoder during training

    mention_embedding: List[float]
        The embedding for the mention text for an example
    positive_concept_emebdding: List[float]
        The embedding for the matching concept text for the mention
    negative_concept_embeddings: List[List[float]]
        The embeddings for disparate concept texts that serve as negative examples.

    """

    mention_embedding: torch.Tensor
    positive_concept_embedding: torch.Tensor
    negative_concept_embeddings: List[torch.Tensor]


class MarginMSELoss(torch.nn.Module):
    """
    MarginMSELoss

    A custom torch loss function that computes margin MSE loss (similar to TripletMarginWithDistanceLoss with a dot product distance but with the ability to support multiple negatives per positive)

    pool_negatives:
        if true, weights averages of all negative scores per example into the weight of one positive example
    """

    def __init__(self, pool_negatives: bool = True):
        super(MarginMSELoss, self).__init__()
        self.pool_negatives = pool_negatives
        # self.mse = torch.nn.MSELoss()

    def avg_loss(self, predictions: Iterable[EncoderPredictions]):
        """
        avg_loss

        Returns the mean marginMSE loss for a list of encoder predictions.

        -------
        PARAMS:

        predictions: Iterable[EncoderPredictions]
            The iterable of predictions made by the encoder as stored in EncoderPrediction namedtuples
        """

        losses = [
            self(
                mention_embedding,
                positive_concept_embedding,
                negative_concept_embedding,
            )
            for mention_embedding, positive_concept_embedding, negative_concept_embedding in predictions
        ]

        return torch.mean(torch.Tensor(losses))

    def forward(
        self,
        prediction: torch.Tensor,
        positive: torch.Tensor,
        negatives: List[torch.Tensor],
    ) -> float:
        """
        forward

        Returns the marginMSE loss function of the encoders over a list of encoder predictions with positive and negative examples.

        More details on marginMSE loss here: https://arxiv.org/pdf/2010.02666.pdf

        -------
        PARAMS:

        predictions: Tensor
            The embedding predicted by the mention encoder model from a mention
        positives: Tensor
            The embedding created by the concept encoder model from the postive concept corresponding to the mention
        negatives: Tensor
            A list of embeddings created by the concept encoder model from the list of negative concepts corresponding to the mention
        """

        target = 2  # Negative weights add up to 1 if they're pooled
        if not self.pool_negatives:
            target = len(negatives) + 1

        positive_score = torch.dot(prediction, positive)

        negative_score = 0
        for negative in negatives:
            if self.pool_negatives:
                negative_score -= torch.dot(prediction, negative) / len(negatives)
            else:
                negative_score -= torch.dot(prediction, negative)

        actual = positive_score + negative_score

        mse = torch.mean(torch.square(target - actual))

        return mse

    def __call__(
        self,
        predictions: Iterable[torch.Tensor],
        positives: Iterable[torch.Tensor],
        negatives: Iterable[List[torch.Tensor]],
    ) -> float:
        # Same function as above, not gonna give it its own docstring
        return self.forward(predictions, positives, negatives)


class EntityLinker:
    """
    This class is responsible for using the encoder on entities and sets of mentions in order to identify the closest entity link to mentions

    lkb: LexicalKnoweledgeBase
        The lexical knowledge base of linkable concepts
    mention_encoder: Encoder
        The encoder for entity mentions
    candidate_encoder: Encoder
        The encoder for candidate names/definitions
    span_key: str
        The key under which spans are stored in SpaCy Doc spans
    candidate_embeddings: Dict[Concept, List[float]]
        The cache of embeddings for all relevant candidate concepts
    min_score: float
        The min dot product score required to link a concept to an entity (below which no entity is assigned)
    separator_token: str
        The token used to separate context from mentions and names from definitions
    """

    def __init__(self, lkb: LexicalKnowledgeBase, span_key: str, min_score: float):
        self.lkb: LexicalKnowledgeBase = lkb
        self.mention_encoder: Encoder = Encoder()
        self.candidate_encoder: Encoder = Encoder()
        self.span_key: str = span_key
        self.candidate_embeddings: Dict[Concept, List[float]] = {}
        self.min_score: float = min_score
        self.separator_token: str = "[SEP]"

        # self.initialize_candidate_embeddings()

    def initialize_candidate_embeddings(self):
        """
        initialize_candidate_embeddings

        Iterate over all the concepts in the stored lkb and cache their embeddings
        """
        for candidate_entity in get_concept_list(self.lkb):
            name = candidate_entity.canonical_name
            definition = candidate_entity.definition
            self.candidate_embeddings[candidate_entity] = self.candidate_encoder(
                name + self.separator_token + definition
            )

    def get_encoder_dot_product_score(
        self, pair: Tuple[str, str], is_similar: bool = True
    ) -> float:
        """
        get_encoder_dot_product_score

        This function measures the similarity between a pair of texts through the mention and candidate encoder using dot product. A negative score is desired for different texts; a postive one for similar texts. This is used in computation of the loss function.

        ------
        PARAMS:

        pair: Tuple[str, str]
            The pair of texts whose embeddings are compaired. The input to the mention encoder should come first, then the candidate encoder.

        is_similar: bool
            Whether the provided pair of texts should be considered similar or disparate. This flips the sign of the dot product score.
        """

        mention_embedding = self.mention_encoder(pair[0])
        candidate_embedding = self.candidate_encoder(pair[1])

        return Distancer.dot_product(mention_embedding, candidate_embedding) * (
            1 if is_similar else -1
        )

    def get_loss(self, examples: Iterable[EncoderPredictions]) -> float:
        """
        get_loss

        Returns the marginMSE loss function of the encoders over a set of encoder predictions. This is NOT a torch compatible loss function.

        More details on marginMSE loss here: https://arxiv.org/pdf/2010.02666.pdf

        -------
        PARAMS:

        examples: Iterable[EncoderPredictions]
            An iterable of predictions produced by the encoder containing embeddings for a mention as well as positive and negative concept embeddings.
        """

        squared_errors = []

        for mention, positive_concept, negative_concepts in examples:
            target = 1 + len(
                negative_concepts
            )  # The ideal dp score is 1 for each trial
            similarity = torch.dot(mention, positive_concept)
            for negative_concept in negative_concepts:
                similarity -= torch.dot(mention, negative_concept)

            squared_errors.append(torch.square(target - similarity))

        mse = torch.mean(torch.Tensor(squared_errors))

        return mse

    def train(
        self,
        training_loader: DataLoader,
        validation_loader: DataLoader,
        num_epochs: int = 10,
        learning_rate: float = 1e-5,  # Recommended by https://arxiv.org/pdf/2103.05028.pdf
        verbose: bool = True,
    ) -> float:
        start_time = datetime.now()
        writer = SummaryWriter(
            "runs/DE_CEL_{}".format(start_time.strftime("%Y%m%d_%H%M%S"))
        )

        # Because we're training 2 models at once, we pool their parameters
        parameters = list(self.candidate_encoder.model.parameters()) + list(
            self.mention_encoder.model.parameters()
        )

        # AdamW optimizer recommended by https://arxiv.org/pdf/2103.05028.pdf
        optimizer = torch.optim.AdamW(parameters, lr=learning_rate)

        loss_fn = MarginMSELoss(pool_negatives=True)

        best_v_loss = 2**31

        for epoch_idx in range(num_epochs):
            if verbose:
                epoch_time = datetime.now()
                print("Now starting epoch {} at {}".format(epoch_idx, epoch_time))

            # Run training for one epoch and report back the avg. loss for the last batch
            avg_loss = self.train_epoch(
                epoch_idx=epoch_idx,
                tb_writer=writer,
                training_loader=training_loader,
                optimizer=optimizer,
                loss_fn=loss_fn,
                batch_size=1000,
                verbose=verbose,
            )

            # Put the encoders into eval mode
            self.candidate_encoder.model.eval()
            self.mention_encoder.model.eval()

            with torch.no_grad():
                v_data_size = 0
                running_v_loss = 0
                for v_data in validation_loader:
                    v_mention, v_postive, v_negatives = v_data

                    v_mention_embedding = self.mention_encoder(v_mention)
                    v_positive_embedding = self.candidate_encoder(v_postive)
                    v_negative_embeddings = [
                        self.candidate_encoder(v_negative) for v_negative in v_negatives
                    ]

                    v_loss = loss_fn(
                        v_mention_embedding,
                        v_positive_embedding,
                        v_negative_embeddings,
                    )
                    running_v_loss += v_loss
                    v_data_size += 1

                avg_v_loss = running_v_loss / (v_data_size + 1)

            if verbose:
                print(
                    "Training loss: {}  Validation loss: {}".format(
                        avg_loss, avg_v_loss
                    )
                )

            writer.add_scalars(
                "Training vs. Validation Loss",
                {"Training": avg_loss, "Validation": avg_v_loss},
                epoch_idx,
            )

            writer.flush()

            # Save the model if it's beat its own record
            if avg_v_loss < best_v_loss:
                best_v_loss = avg_v_loss
                model_path = "models/model_{}_{}".format(start_time, epoch_idx)
                torch.save(
                    self.candidate_encoder.model.state_dict(), model_path + "_candidate"
                )
                torch.save(
                    self.mention_encoder.model.state_dict(), model_path + "_mention"
                )
        if verbose:
            print("Done! At {}".format(datetime.now()))

    def train_epoch(
        self,
        epoch_idx: int,
        tb_writer: Optional[SummaryWriter],
        training_loader: DataLoader,
        optimizer: torch.optim.Optimizer,
        loss_fn: "torch.nn.modules._Loss",
        batch_size: int = 1000,
        verbose: bool = True,
    ):
        """
        train_epoch

        Trains the underlying models in the encoders for one epoch. Returns the average loss for the last batch

        -------
        PARAMS:

        epoch_idx: int
            The number of the current epoch, starting at 0
        tb_writer: Optional[SummaryWriter]
            A SummaryWriter that batch-by-batch stats will be reported to. If None is provided, nothing will be reported.
        training_loader: DataLoader
            A DataLoader that contains the training data for the model. Each element should contain a triplet of a mention to encode, a postive concept text that is similar to it, and a negative concept text that is dissimilar to it
        optimizer: Optimizer
            The models' optimizer; usually set in the main train method
        loss_fn: _Loss
            The models' loss function; usually set in the main train method
        batch_size: int
            The size of each batch; summary stats will be computed and reported after each one has elapsed
        verbose: bool
            Whether to print summary stats to the console
        """

        running_loss = 0.0
        last_loss = 0.0

        for idx, data in enumerate(training_loader):
            # Type: Tuple[str, torch.Tensor, List[torch.Tensor]]
            mention, positive, negatives = data

            optimizer.zero_grad()

            mention_embedding = self.mention_encoder(mention)
            positive_embedding = self.candidate_encoder(positive)

            # There can be multiple negative candidates for each mention
            negative_embeddings = [
                self.candidate_encoder(negative) for negative in negatives
            ]

            loss_value = loss_fn(
                mention_embedding, positive_embedding, negative_embeddings
            )

            loss_value.backward()

            running_loss += loss_value

            if verbose and idx % batch_size == batch_size - 1:
                batch_loss = running_loss / batch_size - 1
                print("\tbatch {} loss: {}".format(idx + 1, batch_loss))
                tb_x = epoch_idx * len(training_loader) + idx + 1
                if tb_writer is not None:
                    tb_writer.add_scalar("Loss/train", last_loss, tb_x)
                running_loss = 0
                last_loss = batch_loss
        return last_loss

    def __call__(
        self,
        doc: Doc,
        similarity_scorer: Callable[[torch.Tensor, torch.Tensor], float] = torch.dot,
    ):
        """
        __call__

        Assigns entity links for all of the spans under the linker's span key

        -----
        PARAMS:

        doc: Doc
            The document which entity links get assigned to
        similarity_scorer:
            A function which takes two torch tensors and returns a score based on their similarity, where a high value indicates high similarity
        """

        for span in doc.spans[self.span_key]:
            span_embedding = self.mention_encoder(
                span.text
                + self.separator_token
                + ContextWindow.get_context_window(
                    span, size=64
                ).get_text()  # A context window size of 128 is used: radius ~= 60
            )
            best_entity = None
            best_similarity = None

            # Iterate over candidate entities to compute the one with the highest similarity score
            for candidate_entity in get_concept_list(self.lkb):
                similarity = similarity_scorer(
                    span_embedding, self.candidate_embeddings[candidate_entity]
                )  # Generally, this is dot product
                if best_similarity is None or similarity > best_similarity:
                    best_similarity = similarity
                    best_entity = candidate_entity

            if best_similarity > self.min_score:
                span._.linked_entity = best_entity
            span._.entity_score = best_similarity
            name = (
                span._.linked_entity.canonical_name
                if span._.linked_entity is not None
                else "None"
            )
            # print("{} linked to {} w/ score {}".format(span.text, name, span._.entity_score))
        return doc


def get_concept_list(lkb: LexicalKnowledgeBase) -> Iterable[Concept]:
    """
    Returns a list of concepts from a LexicalKnowledgeBase class
    """

    return lkb.cui_to_concept.values()
