import json
from typing import Literal, Optional
from spacy.language import Language
from spacy.tokens import Doc

from crtl.components.crtl_el import get_setters_and_getters


class LinkOverRider:
    def __init__(
        self,
        case_sensitive: bool,
        ent_or_span: Literal["ent", "span"],
        span_key: Optional[str] = None,
    ):
        self._case_sensitive = case_sensitive
        self.span_getter, self.span_setter = get_setters_and_getters(
            ent_or_span, span_key
        )

    def initialize(self, get_examples=None, nlp=None, path=None):
        self._load_json(path)

    def to_disk(self, path, exclude=tuple()):
        if not path.exists():
            path.mkdir()
        with open(path / "overrides_path.json", "w+") as f:
            json.dump(self._original_dictionary, f)

    def _load_json(self, path):
        with open(path) as f:
            self._original_dictionary = json.load(f)
        if self._case_sensitive:
            self._overrides_dict = self._original_dictionary
        else:
            self._overrides_dict = {
                key.lower(): value for key, value in self._original_dictionary.items()
            }
        return self

    def from_disk(self, path, exclude=tuple()):
        self._load_json(path / "overrides_path.json")

    def __call__(self, doc: Doc):
        overridden_ents = []
        for ent in self.span_getter(doc):
            # Overwrite the link there's an entry in the data
            if self._case_sensitive:
                ent_text = ent.text
            else:
                ent_text = ent.text.lower()
            if ent_text in self._overrides_dict:
                ent.kb_id_ = self._overrides_dict[ent_text]
            overridden_ents.append(ent)
        self.span_setter(doc, overridden_ents)
        return doc


@Language.factory(
    name="trend_literal_link_overrider",
    default_config={"case_sensitive": False, "ent_or_span": "ent", "span_key": None},
)
def make_ent_link_overrider(
    nlp: Language,
    name: str,
    case_sensitive: bool,
    ent_or_span: str,
    span_key: Optional[str] = None,
):
    return LinkOverRider(case_sensitive, ent_or_span, span_key)
